==> unification-0.0.1.tar.gz: lib/unification.dart <==
import "package:tailcalls/tailcalls.dart";
import 'utils.dart';

/// unification, data structures and functions

/// Logic data structure

abstract class Termtype {}

class Var extends Termtype {
  String _id;
  Var(i) : assert(i != null) {
    if (i == null) {
      this._id = "";
    } else {
      this._id = i;
    }
  }
  String get id => this._id;
  @override
  String toString() {
    return "Var(${this.id})";
  }
}

class Term extends Termtype {
  String _id;
  List<Termtype> _termlist;
  Term(i, t) : assert(i != null && t != null) {
    if (i == null) {
      this._id = "";
    } else {
      this._id = i;
    }

    if (t == null) {
      this._termlist = <Termtype>[];
    } else {
      this._termlist = t;
    }
  }
  List<Termtype> get termlist => this._termlist;
  String get id => this._id;

  @override
  String toString() {
    return "Term(${this.id}, ${this.termlist})";
  }
}

////////////////////
///
/// Utility  Datastructure
///

class Tupl<L, R> {
  L _left;
  R _right;
  Tupl(left, right) : assert(left != null && right != null) {
    if (left == null) {
      throw new Exception("Tupl: left is null");
    } else {
      this._left = left;
    }

    if (right == null) {
      throw new Exception("Tupl: right is null");
    } else {
      this._right = right;
    }
  }

  L get left => this._left;
  R get right => this._right;

  @override
  String toString() {
    return "(${left.toString()}, ${right.toString()})";
  }
}

// functions, related to unification

/// [occurs]: "occurs check"

TailRec<bool> occurs(String x, Termtype t) {
  if (t is Var) {
    var y = t.id;
    return done(x == y);
  } else if (t is Term) {
    var s = t.termlist;
    return tailcall(() => _exsts(s, x));
  } else if (x == null || t == null) {
    throw new Exception("occurs: Variable name  or Termtype is null");
  } else {
    throw new Exception("occurs: Unknown Exception");
  }
}

/// [exsts]: helper fuction for "occurs check"

TailRec<bool> _exsts(List<Termtype> l, String target) {
  if (l.isEmpty) {
    return done(false);
  } else {
    Termtype lh = l.first;
    List<Termtype> lt = l.sublist(1);

    return tailcall(() => _exsts(lt, target)).flatMap((right) {
      return tailcall(() => occurs(target, lh)).map((left) {
        return (left || right);
      });
    });
  }
}

/// [subst]: substitution

TailRec<Termtype> _subst(Termtype s, String x, Termtype t) {
  if (t is Var) {
    if (x == t.id) {
      return done(s);
    } else {
      return done(t);
    }
  } else if (t is Term) {
    String f = t.id;
    List<Termtype> u = t.termlist; // XTODO: u ist null
    //assert(! t.t.isEmpty);// XTODO: fails
    assert(u != null); // XTODO: fails
    return tailcall(() => _mp(s, x, u)).map((right) {
      return new Term(/*t.i*/ f, right);
    });

    // for
    //
    // var right = tailcall(mp(s, x, u)); //<-
    // return new Term(t.i, right); // yield
    //
  } else {
    throw new Exception("Subst: Unbehandelter Fall");
  }
}

/// [mp]

TailRec<List<Termtype>> _mp(Termtype s, String x, List<Termtype> l) {
  assert(l != null); // XTODO
  if (l.isEmpty) {
    return done<List<Termtype>>(new List<Termtype>());
  } else {
    var lh = l.first;
    var lt = l.sublist(1);
    assert(lh != null);
    assert(lt != null);
    // for
    return tailcall(() => _mp(s, x, lt)).flatMap((right) {
      assert(lh != null);
      return tailcall(() => _subst(s, x, lh)).map((left) {
        assert(right != null);
        assert(left != null);
        var res = [left];
        res.addAll(right); // XTODO
        return res;
      });
    });

    // for
    //
    // var right = tailcall(mp(s, x, lt)); //<-
    // var left = tailcall(subst(s, x, lh)); //<-
    // [left].addAll(right);
    // return left; //yield
    //
  }
}

/// [apply]

TailRec<Termtype> _apply(List<Tupl<String, Termtype>> ths, Termtype z) {
  if (ths.isEmpty) {
    return done<Termtype>(z);
  } else {
    Tupl<String, Termtype> frst = ths.first;
    String x = frst.left;
    Termtype u = frst.right;

    var xs = ths.sublist(1); // XTODO

    // for
    return tailcall(() => _apply(xs, z)).flatMap((apd) {
      assert(apd != null);
      if (apd is Term) {
        assert(apd.id != null);
      }
      return tailcall(() => _subst(u, x, apd)).map((right) {
        return right;
      });
    });

    // for
    //
    // var apd = tailcall(()=>apply(xs, z)); // <-
    // var right = tailcall(()=>subst(u, x, apd)); //<-
    // return right; // yield
    //

  }
}

/// [unify_one]: unify two ingle Termtypes, one by one

TailRec<List<Tupl<String, Termtype>>> _unify_one(Termtype s, Termtype t) {
  if (s == null || t == null) {
    throw new Exception("occurs: Termtype is null");
  } else if (s is Var && t is Var) {
    var x = s.id;
    var y = t.id;

    if (x == y) {
      return done<List<Tupl<String, Termtype>>>(
          new List<Tupl<String, Termtype>>());
    } else {
      return done<List<Tupl<String, Termtype>>>(
          new List<Tupl<String, Termtype>>()
            ..add(new Tupl<String, Termtype>(x, t)));
    }
  } else if (s is Term && t is Term) {
    var f = s.id;
    var sc = s.termlist;

    var g = t.id;
    var tc = t.termlist;

    if ((f == g) && (sc.length == tc.length)) {
      List<Tupl<Termtype, Termtype>> zpd =
          zip<Termtype, Tupl<Termtype, Termtype>>(
              sc, tc, (left, right) => new Tupl(left, right));

      return tailcall(() => unify(zpd));
    } else {
      throw new Exception("Not unifiable #1");
    }
  } else if (s is Var && t is Term) {
    // case (Var(x), t @ Term(_, _)) => help3(t: termtype, x:id)
    return _unifyhelper(t, s.id);

    /*
    String x = s.i;
    List<Tupl<String, Termtype>> res1 = new List<Tupl<String, Termtype>>()
      ..add(new Tupl<String, Termtype>(x, t));
    return done<List<Tupl<String, Termtype>>>(res1);
    */

  } else if (s is Term && t is Var) {
    // case (t @ Term(_, _), Var(x)) => help3(t: termtype, x:id)
    return _unifyhelper(s, t.id);

    /* 
    // original
    String x = t.i;
    List<Tupl<String, Termtype>> res2 = new List<Tupl<String, Termtype>>()
      ..add(new Tupl<String, Termtype>(x, t));
    return done<List<Tupl<String, Termtype>>>(res2);
    */

  } else {
    throw new Exception("Not unifiable #2");
  }
}

/// [unifyhelper]: helper function for [unify_one]

TailRec<List<Tupl<String, Termtype>>> _unifyhelper(Termtype t, String x) {
  return tailcall(() => occurs(x, t)).flatMap((left) {
    return done([new Tupl(x, t)]).map((right) {
      List<Tupl<String, Termtype>> innerres;

      if (left) {
        // false when exception
        throw new Exception("not unifiable: circularity");
      } else {
        innerres = right;
      }

      return innerres;
    });
  });
}

/// [unify]: unify a list of terms

TailRec<List<Tupl<String, Termtype>>> unify(List<Tupl<Termtype, Termtype>> s) {
  if (s.isEmpty) {
    return done<List<Tupl<String, Termtype>>>(
        new List<Tupl<String, Termtype>>());
  } else {
    Termtype x = s.first.left;
    Termtype y = s.first.right;
    List<Tupl<Termtype, Termtype>> t = s.sublist(1); // xTODO

    //assert(t  != null); // xTODO

    return tailcall(() => unify(t)).flatMap((t2) {
      // Argument: List<Tupl<String, Termtype>>
      return tailcall(() => _apply(t2, x)).flatMap((left) {
        //Argument:  Termtype
        return tailcall(() => _apply(t2, y)).flatMap((right) {
          //Argument:  Termtype
          return tailcall(() => _unify_one(left, right)).map((t1) {
            //Argument:  List<Tupl<String, Termtype>>
            //assert(t1 != null);
            //assert(t2 != null);
            t1.addAll(t2);
            return t1;
          });
        });
      });
    });

    //
    // for
    //
    // var t2 = tailcall(unify(t)); // <-
    // var left = tailcall(apply(t2, x)); // <-
    // var right = tailcall(apply(t2, y)); // <-
    // var t1 = tailcall(unify_one(left, right)); // <-
    // return t1.addAll(t2);
    //

  }
}

==> unification-0.1.2.tar.gz: lib/unification.dart <==
import "package:tailcalls/tailcalls.dart";
import 'utils.dart';
import 'terms.dart';
export 'terms.dart';

/// trampolined version

/// occurs check

TailRec<bool> occurs(String x, Termtype t) {
  if (t is Var) {
    var y = t.id;
    return done(x == y);
  } else if (t is Term) {
    var s = t.termlist;
    return tailcall(() => _exsts(s, x));
  } else if (x == null || t == null) {
    throw new Exception("occurs: Variable name or Term is null");
  } else {
    throw new Exception("occurs: Unknown Exception");
  }
}

/// helper fuction for occurs check

TailRec<bool> _exsts(List<Termtype> l, String target) {
  if (l.isEmpty) {
    return done(false);
  } else {
    Termtype lh = l.first;
    List<Termtype> lt = l.sublist(1);

    return tailcall(() => _exsts(lt, target)).flatMap((right) {
      return tailcall(() => occurs(target, lh)).map((left) {
        return (left || right);
      });
    });
  }
}

/// substitution

TailRec<Termtype> _subst(Termtype s, String x, Termtype t) {
  if (t is Var) {
    if (x == t.id) {
      return done(s);
    } else {
      return done(t);
    }
  } else if (t is Term) {
    String f = t.id;
    List<Termtype> u = t.termlist;
    assert(!u.isEmpty);
    assert(u != null);
    return tailcall(() => _mp(s, x, u)).map((right) {
      return new Term(/*t.i*/ f, right);
    });
  } else {
    throw new Exception("Subst: Unbehandelter Fall");
  }
}

///

TailRec<List<Termtype>> _mp(Termtype s, String x, List<Termtype> l) {
  assert(l != null);
  if (l.isEmpty) {
    return done<List<Termtype>>(new List<Termtype>());
  } else {
    var lh = l.first;
    var lt = l.sublist(1);
    assert(lh != null);
    assert(lt != null);
    return tailcall(() => _mp(s, x, lt)).flatMap((right) {
      assert(lh != null);
      return tailcall(() => _subst(s, x, lh)).map((left) {
        assert(right != null);
        assert(left != null);
        var res = [left];
        res.addAll(right);
        return res;
      });
    });
  }
}

/// apply substitution

TailRec<Termtype> _apply(List<Tupl<String, Termtype>> ths, Termtype z) {
  if (ths.isEmpty) {
    return done<Termtype>(z);
  } else {
    // pop
    Tupl<String, Termtype> first = ths.first;
    String x = first.left;
    Termtype u = first.right;

    var xs = ths.sublist(1);

    return tailcall(() => _apply(xs, z)).flatMap((Termtype apd) {
      assert(apd != null);
      assert(apd.id != null);

      return tailcall(() => _subst(u, x, apd)).map((right) {
        return right;
      });
    });
  }
}

/// unifies two ingle Termtypes, one by one

TailRec<List<Tupl<String, Termtype>>> _unify_one(Termtype s, Termtype t) {
  if (s == null || t == null) {
    throw new Exception("occurs: Termtype is null");
  } else if (s is Var && t is Var) {
    var x = s.id;
    var y = t.id;

    if (x == y) {
      return done<List<Tupl<String, Termtype>>>(
          new List<Tupl<String, Termtype>>());
    } else {
      return done<List<Tupl<String, Termtype>>>(
          new List<Tupl<String, Termtype>>()
            ..add(new Tupl<String, Termtype>(x, t)));
    }
  } else if (s is Term && t is Term) {
    var f = s.id;
    var sc = s.termlist;

    var g = t.id;
    var tc = t.termlist;

    if ((f == g) && (sc.length == tc.length)) {
      List<Tupl<Termtype, Termtype>> zpd =
          zip<Termtype, Tupl<Termtype, Termtype>>(
              sc, tc, (left, right) => new Tupl(left, right));

      return tailcall(() => unify(zpd));
    } else {
      throw new Exception("Not unifiable #1");
    }
  } else if (s is Var && t is Term) {
    return _unifyhelper(t, s.id);

  } else if (s is Term && t is Var) {
    return _unifyhelper(s, t.id);
  } else {
    throw new Exception("Not unifiable #2");
  }
}

/// helper function for [unify_one]

TailRec<List<Tupl<String, Termtype>>> _unifyhelper(Termtype t, String x) {
  return tailcall(() => occurs(x, t)).flatMap((left) {
    return done([new Tupl(x, t)]).map((right) {
      List<Tupl<String, Termtype>> innerres;
      if (left) {
        throw new Exception("Not unifiable: Circularity");
      } else {
        innerres = right;
      }
      return innerres;
    });
  });
}

/// unify a list of terms

TailRec<List<Tupl<String, Termtype>>> unify(List<Tupl<Termtype, Termtype>> s) {
  if (s.isEmpty) {
    return done<List<Tupl<String, Termtype>>>(
        new List<Tupl<String, Termtype>>());
  } else {
    Termtype x = s.first.left;
    Termtype y = s.first.right;
    List<Tupl<Termtype, Termtype>> t = s.sublist(1);

    assert(t != null);

    return tailcall(() => unify(t)).flatMap((t2) {
      return tailcall(() => _apply(t2, x)).flatMap((left) {
        return tailcall(() => _apply(t2, y)).flatMap((right) {
          return tailcall(() => _unify_one(left, right)).map((t1) {
            assert(t1 != null);
            assert(t2 != null);
            t1.addAll(t2);
            return t1;
          });
        });
      });
    });
  }
}
